'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _regeneratorRuntime = _interopDefault(require('babel-runtime/regenerator'));
var _asyncToGenerator = _interopDefault(require('babel-runtime/helpers/asyncToGenerator'));
var path = _interopDefault(require('path'));
var pify = _interopDefault(require('pify'));
var managers_js = require('./managers.js');
var utils_js = require('./utils.js');
var _getIterator = _interopDefault(require('babel-runtime/core-js/get-iterator'));
var _Number$isNaN = _interopDefault(require('babel-runtime/core-js/number/is-nan'));
var stream = _interopDefault(require('stream'));
var thru = _interopDefault(require('thru'));
var _Map = _interopDefault(require('babel-runtime/core-js/map'));
var _Promise = _interopDefault(require('babel-runtime/core-js/promise'));
var buffer = require('buffer');
var listpack = _interopDefault(require('git-list-pack'));
var peek = _interopDefault(require('buffer-peek-stream'));
var applyDelta = _interopDefault(require('git-apply-delta'));
var models_js = require('./models.js');
var _toConsumableArray = _interopDefault(require('babel-runtime/helpers/toConsumableArray'));
var _Set = _interopDefault(require('babel-runtime/core-js/set'));
var pad = _interopDefault(require('pad'));
var pako = _interopDefault(require('pako'));
var crypto = _interopDefault(require('crypto'));
var _extends = _interopDefault(require('babel-runtime/helpers/extends'));
var openpgp_min_js = require('openpgp/dist/openpgp.min.js');

var add = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref) {
    var gitdir = _ref.gitdir,
        workdir = _ref.workdir,
        filepath = _ref.filepath;
    var type, object, oid;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            type = 'blob';
            _context2.next = 3;
            return utils_js.read(path.join(workdir, filepath));

          case 3:
            object = _context2.sent;

            if (!(object === null)) {
              _context2.next = 6;
              break;
            }

            throw new Error('Could not read file \'' + filepath + '\'');

          case 6:
            _context2.next = 8;
            return managers_js.GitObjectManager.write({ gitdir: gitdir, type: type, object: object });

          case 8:
            oid = _context2.sent;
            _context2.next = 11;
            return managers_js.GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(index) {
                var stats;
                return _regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        _context.next = 2;
                        return pify(utils_js.fs().lstat)(path.join(workdir, filepath));

                      case 2:
                        stats = _context.sent;

                        index.insert({ filepath: filepath, stats: stats, oid: oid });

                      case 4:
                      case 'end':
                        return _context.stop();
                    }
                  }
                }, _callee, this);
              }));

              return function (_x2) {
                return _ref3.apply(this, arguments);
              };
            }());

          case 11:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function add(_x) {
    return _ref2.apply(this, arguments);
  };
}();

// @flow
var init = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref /*: { gitdir: string } */) {
    var gitdir = _ref.gitdir;
    var folders;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            folders = ['hooks', 'info', 'objects/info', 'objects/pack', 'refs/heads', 'refs/tags'];

            folders = folders.map(function (dir) {
              return gitdir + '/' + dir;
            });
            _context.next = 4;
            return utils_js.mkdirs(folders);

          case 4:
            _context.next = 6;
            return utils_js.write(gitdir + '/config', '[core]\n' + '\trepositoryformatversion = 0\n' + '\tfilemode = false\n' + '\tbare = false\n' + '\tlogallrefupdates = true\n' + '\tsymlinks = false\n' + '\tignorecase = true\n');

          case 6:
            _context.next = 8;
            return utils_js.write(gitdir + '/HEAD', 'ref: refs/heads/master\n');

          case 8:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function init(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var config = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(args) {
    var gitdir, path$$1, value, config, _value;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            gitdir = args.gitdir, path$$1 = args.path, value = args.value;
            _context.next = 3;
            return managers_js.GitConfigManager.get({ gitdir: gitdir });

          case 3:
            config = _context.sent;

            if (!(value === undefined && !args.hasOwnProperty('value'))) {
              _context.next = 11;
              break;
            }

            _context.next = 7;
            return config.get(path$$1);

          case 7:
            _value = _context.sent;
            return _context.abrupt('return', _value);

          case 11:
            _context.next = 13;
            return config.set(path$$1, value);

          case 13:
            _context.next = 15;
            return managers_js.GitConfigManager.save({ gitdir: gitdir, config: config });

          case 15:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function config(_x) {
    return _ref.apply(this, arguments);
  };
}();

// @flow
/*::
import type {Writable} from 'stream'
*/

var types = {
  1: 'commit',
  2: 'tree',
  3: 'blob',
  4: 'tag',
  6: 'ofs-delta',
  7: 'ref-delta'
};

function parseVarInt(buffer$$1 /*: Buffer */) {
  var n = 0;
  for (var i = 0; i < buffer$$1.byteLength; i++) {
    n = (buffer$$1[i] & 127) + (n << 7);
    if ((buffer$$1[i] & 128) === 0) {
      if (i !== buffer$$1.byteLength - 1) throw new Error('Invalid varint buffer');
      return n;
    }
  }
  throw new Error('Invalid varint buffer');
}

// TODO: Move this to 'plumbing'
var unpack = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref /*: {gitdir: string, inputStream: ReadableStream, onprogress: Function} */
  ) {
    var gitdir = _ref.gitdir,
        inputStream = _ref.inputStream,
        onprogress = _ref.onprogress;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            return _context2.abrupt('return', new _Promise(function (resolve, reject) {
              var _this = this;

              // Read header
              peek(inputStream, 12, function (err, data, inputStream) {
                if (err) return reject(err);
                var iden = data.slice(0, 4).toString('utf8');
                if (iden !== 'PACK') {
                  throw new Error('Packfile started with \'' + iden + '\'. Expected \'PACK\'');
                }
                var ver = data.slice(4, 8).toString('hex');
                if (ver !== '00000002') {
                  throw new Error('Unknown packfile version \'' + ver + '\'. Expected 00000002.');
                }
                // Read a 4 byte (32-bit) int
                var numObjects = data.readInt32BE(8);
                if (onprogress !== undefined) {
                  onprogress({ loaded: 0, total: numObjects, lengthComputable: true });
                }
                if (numObjects === 0) return resolve();
                // And on our merry way
                var offsetMap = new _Map();
                inputStream.pipe(listpack()).pipe(thru(function () {
                  var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref3, next) {
                    var data = _ref3.data,
                        type = _ref3.type,
                        reference = _ref3.reference,
                        offset = _ref3.offset,
                        num = _ref3.num;

                    var oid, _ref5, object, _type, result, newoid, absoluteOffset, referenceOid, _ref6, _type2, _object, _result, _oid, _oid2;

                    return _regeneratorRuntime.wrap(function _callee$(_context) {
                      while (1) {
                        switch (_context.prev = _context.next) {
                          case 0:
                            type = types[type];

                            if (!(type === 'ref-delta')) {
                              _context.next = 21;
                              break;
                            }

                            oid = buffer.Buffer.from(reference).toString('hex');
                            _context.prev = 3;
                            _context.next = 6;
                            return managers_js.GitObjectManager.read({
                              gitdir: gitdir,
                              oid: oid
                            });

                          case 6:
                            _ref5 = _context.sent;
                            object = _ref5.object;
                            _type = _ref5.type;
                            result = applyDelta(data, object);
                            _context.next = 12;
                            return managers_js.GitObjectManager.write({
                              gitdir: gitdir,
                              type: _type,
                              object: result
                            });

                          case 12:
                            newoid = _context.sent;

                            // console.log(`${type} ${newoid} ref-delta ${oid}`)
                            offsetMap.set(offset, newoid);
                            _context.next = 19;
                            break;

                          case 16:
                            _context.prev = 16;
                            _context.t0 = _context['catch'](3);
                            throw new Error('Could not find object ' + reference + ' ' + oid + ' that is referenced by a ref-delta object in packfile at byte offset ' + offset + '.');

                          case 19:
                            _context.next = 40;
                            break;

                          case 21:
                            if (!(type === 'ofs-delta')) {
                              _context.next = 36;
                              break;
                            }

                            // Note: this might be not working because offsets might not be
                            // guaranteed to be on object boundaries? In which case we'd need
                            // to write the packfile to disk first, I think.
                            // For now I've "solved" it by simply not advertising ofs-delta as a capability
                            // during the HTTP request, so Github will only send ref-deltas not ofs-deltas.
                            absoluteOffset = offset - parseVarInt(reference);
                            referenceOid = offsetMap.get(absoluteOffset);
                            // console.log(`${offset} ofs-delta ${absoluteOffset} ${referenceOid}`)

                            _context.next = 26;
                            return managers_js.GitObjectManager.read({
                              gitdir: gitdir,
                              oid: referenceOid
                            });

                          case 26:
                            _ref6 = _context.sent;
                            _type2 = _ref6.type;
                            _object = _ref6.object;
                            _result = applyDelta(data, _object);
                            _context.next = 32;
                            return managers_js.GitObjectManager.write({
                              gitdir: gitdir,
                              type: _type2,
                              object: _result
                            });

                          case 32:
                            _oid = _context.sent;

                            // console.log(`${offset} ${type} ${oid} ofs-delta ${referenceOid}`)
                            offsetMap.set(offset, _oid);
                            _context.next = 40;
                            break;

                          case 36:
                            _context.next = 38;
                            return managers_js.GitObjectManager.write({
                              gitdir: gitdir,
                              type: type,
                              object: data
                            });

                          case 38:
                            _oid2 = _context.sent;

                            // console.log(`${offset} ${type} ${oid}`)
                            offsetMap.set(offset, _oid2);

                          case 40:
                            if (onprogress !== undefined) {
                              onprogress({
                                loaded: numObjects - num,
                                total: numObjects,
                                lengthComputable: true
                              });
                            }

                            if (!(num === 0)) {
                              _context.next = 43;
                              break;
                            }

                            return _context.abrupt('return', resolve());

                          case 43:
                            next(null);

                          case 44:
                          case 'end':
                            return _context.stop();
                        }
                      }
                    }, _callee, _this, [[3, 16]]);
                  }));

                  return function (_x2, _x3) {
                    return _ref4.apply(this, arguments);
                  };
                }())).on('error', reject).on('finish', resolve);
              });
            }));

          case 1:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function unpack(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var resolveRef = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        ref = _ref.ref,
        depth = _ref.depth;
    var sha;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            if (!(depth !== undefined)) {
              _context.next = 4;
              break;
            }

            depth--;

            if (!(depth === -1)) {
              _context.next = 4;
              break;
            }

            return _context.abrupt('return', ref);

          case 4:
            sha = void 0;
            // Is it a ref pointer?

            if (!ref.startsWith('ref: ')) {
              _context.next = 8;
              break;
            }

            ref = ref.slice('ref: '.length);
            return _context.abrupt('return', resolveRef({ gitdir: gitdir, ref: ref, depth: depth }));

          case 8:
            if (!(ref.length === 40 && /[0-9a-f]{40}/.test(ref))) {
              _context.next = 10;
              break;
            }

            return _context.abrupt('return', ref);

          case 10:
            if (!(ref === 'HEAD' || ref === 'MERGE_HEAD')) {
              _context.next = 16;
              break;
            }

            _context.next = 13;
            return utils_js.read(gitdir + '/' + ref, { encoding: 'utf8' });

          case 13:
            sha = _context.sent;

            if (!sha) {
              _context.next = 16;
              break;
            }

            return _context.abrupt('return', resolveRef({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

          case 16:
            if (!ref.startsWith('refs/')) {
              _context.next = 22;
              break;
            }

            _context.next = 19;
            return utils_js.read(gitdir + '/' + ref, { encoding: 'utf8' });

          case 19:
            sha = _context.sent;

            if (!sha) {
              _context.next = 22;
              break;
            }

            return _context.abrupt('return', resolveRef({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

          case 22:
            _context.next = 24;
            return utils_js.read(gitdir + '/refs/heads/' + ref, { encoding: 'utf8' });

          case 24:
            sha = _context.sent;

            if (!sha) {
              _context.next = 27;
              break;
            }

            return _context.abrupt('return', resolveRef({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

          case 27:
            _context.next = 29;
            return utils_js.read(gitdir + '/refs/tags/' + ref, { encoding: 'utf8' });

          case 29:
            sha = _context.sent;

            if (!sha) {
              _context.next = 32;
              break;
            }

            return _context.abrupt('return', resolveRef({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

          case 32:
            _context.next = 34;
            return utils_js.read(gitdir + '/refs/remotes/' + ref, { encoding: 'utf8' });

          case 34:
            sha = _context.sent;

            if (!sha) {
              _context.next = 37;
              break;
            }

            return _context.abrupt('return', resolveRef({ gitdir: gitdir, ref: sha.trim(), depth: depth }));

          case 37:
            throw new Error('Could not resolve reference ' + ref);

          case 38:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function resolveRef(_x) {
    return _ref2.apply(this, arguments);
  };
}();

// @flow
var fetchPackfile = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref) {
    var _this = this;

    var gitdir = _ref.gitdir,
        ref = _ref.ref,
        remote = _ref.remote,
        url = _ref.url,
        authUsername = _ref.authUsername,
        authPassword = _ref.authPassword,
        _ref$depth = _ref.depth,
        depth = _ref$depth === undefined ? null : _ref$depth,
        _ref$since = _ref.since,
        since = _ref$since === undefined ? null : _ref$since,
        _ref$exclude = _ref.exclude,
        exclude = _ref$exclude === undefined ? [] : _ref$exclude,
        _ref$relative = _ref.relative,
        relative = _ref$relative === undefined ? false : _ref$relative;

    var remoteHTTP, want, capabilities, packstream, oids, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, oid, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, x, have, response;

    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            if (!(depth !== null)) {
              _context2.next = 4;
              break;
            }

            if (!_Number$isNaN(parseInt(depth))) {
              _context2.next = 3;
              break;
            }

            throw new Error('Invalid value for depth argument: ' + depth);

          case 3:
            depth = parseInt(depth);

          case 4:
            remote = remote || 'origin';

            if (!(url === undefined)) {
              _context2.next = 9;
              break;
            }

            _context2.next = 8;
            return config({
              gitdir: gitdir,
              path: 'remote.' + remote + '.url'
            });

          case 8:
            url = _context2.sent;

          case 9:
            remoteHTTP = new managers_js.GitRemoteHTTP(url);

            if (authUsername !== undefined && authPassword !== undefined) {
              remoteHTTP.auth = {
                username: authUsername,
                password: authPassword
              };
            }
            _context2.next = 13;
            return remoteHTTP.preparePull();

          case 13:
            if (!(depth !== null && !remoteHTTP.capabilities.has('shallow'))) {
              _context2.next = 15;
              break;
            }

            throw new Error('Remote does not support shallow fetches');

          case 15:
            if (!(since !== null && !remoteHTTP.capabilities.has('deepen-since'))) {
              _context2.next = 17;
              break;
            }

            throw new Error('Remote does not support shallow fetches by date');

          case 17:
            if (!(exclude.length > 0 && !remoteHTTP.capabilities.has('deepen-not'))) {
              _context2.next = 19;
              break;
            }

            throw new Error('Remote does not support shallow fetches excluding commits reachable by refs');

          case 19:
            if (!(relative === true && !remoteHTTP.capabilities.has('deepen-relative'))) {
              _context2.next = 21;
              break;
            }

            throw new Error('Remote does not support shallow fetches relative to the current shallow depth');

          case 21:
            _context2.next = 23;
            return managers_js.GitRefsManager.updateRemoteRefs({
              gitdir: gitdir,
              remote: remote,
              refs: remoteHTTP.refs,
              symrefs: remoteHTTP.symrefs
            });

          case 23:
            _context2.next = 25;
            return resolveRef({ gitdir: gitdir, ref: 'refs/remotes/' + remote + '/' + ref });

          case 25:
            want = _context2.sent;

            // Note: I removed "ofs-delta" from the capabilities list and now
            // Github uses all ref-deltas when I fetch packfiles instead of all ofs-deltas. Nice!
            capabilities = 'multi_ack_detailed no-done side-band-64k thin-pack agent=git/' + utils_js.pkg.name + '@' + utils_js.pkg.version + (relative ? ' deepen-relative' : '');
            packstream = new stream.PassThrough();

            packstream.write(models_js.GitPktLine.encode('want ' + want + ' ' + capabilities + '\n'));
            _context2.next = 31;
            return managers_js.GitShallowManager.read({ gitdir: gitdir });

          case 31:
            oids = _context2.sent;

            if (!(oids.size > 0 && remoteHTTP.capabilities.has('shallow'))) {
              _context2.next = 52;
              break;
            }

            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context2.prev = 36;

            for (_iterator = _getIterator(oids); !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
              oid = _step.value;

              packstream.write(models_js.GitPktLine.encode('shallow ' + oid + '\n'));
            }
            _context2.next = 44;
            break;

          case 40:
            _context2.prev = 40;
            _context2.t0 = _context2['catch'](36);
            _didIteratorError = true;
            _iteratorError = _context2.t0;

          case 44:
            _context2.prev = 44;
            _context2.prev = 45;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 47:
            _context2.prev = 47;

            if (!_didIteratorError) {
              _context2.next = 50;
              break;
            }

            throw _iteratorError;

          case 50:
            return _context2.finish(47);

          case 51:
            return _context2.finish(44);

          case 52:
            if (depth !== null) {
              packstream.write(models_js.GitPktLine.encode('deepen ' + depth + '\n'));
            }
            if (since !== null) {
              packstream.write(models_js.GitPktLine.encode('deepen-since ' + Math.floor(since.valueOf() / 1000) + '\n'));
            }
            _iteratorNormalCompletion2 = true;
            _didIteratorError2 = false;
            _iteratorError2 = undefined;
            _context2.prev = 57;
            for (_iterator2 = _getIterator(exclude); !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              x = _step2.value;

              packstream.write(models_js.GitPktLine.encode('deepen-not ' + x + '\n'));
            }
            _context2.next = 65;
            break;

          case 61:
            _context2.prev = 61;
            _context2.t1 = _context2['catch'](57);
            _didIteratorError2 = true;
            _iteratorError2 = _context2.t1;

          case 65:
            _context2.prev = 65;
            _context2.prev = 66;

            if (!_iteratorNormalCompletion2 && _iterator2.return) {
              _iterator2.return();
            }

          case 68:
            _context2.prev = 68;

            if (!_didIteratorError2) {
              _context2.next = 71;
              break;
            }

            throw _iteratorError2;

          case 71:
            return _context2.finish(68);

          case 72:
            return _context2.finish(65);

          case 73:
            packstream.write(models_js.GitPktLine.flush());
            have = null;
            _context2.prev = 75;
            _context2.next = 78;
            return resolveRef({ gitdir: gitdir, ref: ref });

          case 78:
            have = _context2.sent;
            _context2.next = 83;
            break;

          case 81:
            _context2.prev = 81;
            _context2.t2 = _context2['catch'](75);

          case 83:
            if (have) {
              packstream.write(models_js.GitPktLine.encode('have ' + have + '\n'));
              packstream.write(models_js.GitPktLine.flush());
            }
            packstream.end(models_js.GitPktLine.encode('done\n'));
            _context2.next = 87;
            return remoteHTTP.pull(packstream);

          case 87:
            response = _context2.sent;

            response.packetlines.pipe(thru(function () {
              var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(data, next) {
                var line, _oid, _oid2;

                return _regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        line = data.toString('utf8');

                        if (!line.startsWith('shallow')) {
                          _context.next = 10;
                          break;
                        }

                        _oid = line.slice(-41).trim();

                        if (!(_oid.length !== 40)) {
                          _context.next = 5;
                          break;
                        }

                        throw new Error('non-40 character \'shallow\' oid: ' + _oid);

                      case 5:
                        oids.add(_oid);
                        _context.next = 8;
                        return managers_js.GitShallowManager.write({ gitdir: gitdir, oids: oids });

                      case 8:
                        _context.next = 17;
                        break;

                      case 10:
                        if (!line.startsWith('unshallow')) {
                          _context.next = 17;
                          break;
                        }

                        _oid2 = line.slice(-41).trim();

                        if (!(_oid2.length !== 40)) {
                          _context.next = 14;
                          break;
                        }

                        throw new Error('non-40 character \'shallow\' oid: ' + _oid2);

                      case 14:
                        oids.delete(_oid2);
                        _context.next = 17;
                        return managers_js.GitShallowManager.write({ gitdir: gitdir, oids: oids });

                      case 17:
                        next(null, data);

                      case 18:
                      case 'end':
                        return _context.stop();
                    }
                  }
                }, _callee, _this);
              }));

              return function (_x2, _x3) {
                return _ref3.apply(this, arguments);
              };
            }()));
            return _context2.abrupt('return', response);

          case 90:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this, [[36, 40, 44, 52], [45,, 47, 51], [57, 61, 65, 73], [66,, 68, 72], [75, 81]]);
  }));

  return function fetchPackfile(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var fetch = function () {
  var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(_ref4) {
    var gitdir = _ref4.gitdir,
        _ref4$ref = _ref4.ref,
        ref = _ref4$ref === undefined ? 'HEAD' : _ref4$ref,
        remote = _ref4.remote,
        url = _ref4.url,
        authUsername = _ref4.authUsername,
        authPassword = _ref4.authPassword,
        depth = _ref4.depth,
        since = _ref4.since,
        exclude = _ref4.exclude,
        relative = _ref4.relative,
        onprogress = _ref4.onprogress;
    var response;
    return _regeneratorRuntime.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            _context3.next = 2;
            return fetchPackfile({
              gitdir: gitdir,
              ref: ref,
              remote: remote,
              url: url,
              authUsername: authUsername,
              authPassword: authPassword,
              depth: depth,
              since: since,
              exclude: exclude,
              relative: relative
            });

          case 2:
            response = _context3.sent;
            _context3.next = 5;
            return unpack({ gitdir: gitdir, inputStream: response.packfile, onprogress: onprogress });

          case 5:
          case 'end':
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));

  return function fetch(_x4) {
    return _ref5.apply(this, arguments);
  };
}();

var writeTreeToDisk = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        workdir = _ref.workdir,
        index = _ref.index,
        prefix = _ref.prefix,
        tree = _ref.tree;

    var _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, entry, _ref3, type, object, entrypath, filepath, stats, _tree;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context.prev = 3;
            _iterator = _getIterator(tree);

          case 5:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context.next = 32;
              break;
            }

            entry = _step.value;
            _context.next = 9;
            return managers_js.GitObjectManager.read({
              gitdir: gitdir,
              oid: entry.oid
            });

          case 9:
            _ref3 = _context.sent;
            type = _ref3.type;
            object = _ref3.object;
            entrypath = prefix === '' ? entry.path : prefix + '/' + entry.path;
            filepath = path.join(workdir, prefix, entry.path);
            _context.t0 = type;
            _context.next = _context.t0 === 'blob' ? 17 : _context.t0 === 'tree' ? 24 : 28;
            break;

          case 17:
            _context.next = 19;
            return utils_js.write(filepath, object);

          case 19:
            _context.next = 21;
            return pify(utils_js.fs().lstat)(filepath);

          case 21:
            stats = _context.sent;

            index.insert({
              filepath: entrypath,
              stats: stats,
              oid: entry.oid
            });
            return _context.abrupt('break', 29);

          case 24:
            _tree = models_js.GitTree.from(object);
            _context.next = 27;
            return writeTreeToDisk({
              gitdir: gitdir,
              workdir: workdir,
              index: index,
              prefix: entrypath,
              tree: _tree
            });

          case 27:
            return _context.abrupt('break', 29);

          case 28:
            throw new Error('Unexpected object type ' + type + ' found in tree for \'' + entrypath + '\'');

          case 29:
            _iteratorNormalCompletion = true;
            _context.next = 5;
            break;

          case 32:
            _context.next = 38;
            break;

          case 34:
            _context.prev = 34;
            _context.t1 = _context['catch'](3);
            _didIteratorError = true;
            _iteratorError = _context.t1;

          case 38:
            _context.prev = 38;
            _context.prev = 39;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 41:
            _context.prev = 41;

            if (!_didIteratorError) {
              _context.next = 44;
              break;
            }

            throw _iteratorError;

          case 44:
            return _context.finish(41);

          case 45:
            return _context.finish(38);

          case 46:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[3, 34, 38, 46], [39,, 41, 45]]);
  }));

  return function writeTreeToDisk(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var checkout = function () {
  var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(_ref4) {
    var workdir = _ref4.workdir,
        gitdir = _ref4.gitdir,
        remote = _ref4.remote,
        ref = _ref4.ref;

    var oid, remoteRef, commit, comm, sha, _ref6, type, object, tree;

    return _regeneratorRuntime.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            // Get tree oid
            oid = void 0;

            if (!remote) {
              _context3.next = 18;
              break;
            }

            remoteRef = void 0;

            if (!(ref === undefined)) {
              _context3.next = 10;
              break;
            }

            _context3.next = 6;
            return resolveRef({ gitdir: gitdir, ref: remote + '/HEAD', depth: 2 });

          case 6:
            remoteRef = _context3.sent;

            ref = path.basename(remoteRef);
            _context3.next = 11;
            break;

          case 10:
            remoteRef = remote + '/' + ref;

          case 11:
            _context3.next = 13;
            return resolveRef({ gitdir: gitdir, ref: remoteRef });

          case 13:
            oid = _context3.sent;
            _context3.next = 16;
            return utils_js.write(gitdir + '/refs/heads/' + ref, oid + '\n');

          case 16:
            _context3.next = 23;
            break;

          case 18:
            if (!(ref === undefined)) {
              _context3.next = 20;
              break;
            }

            throw new Error('Cannot checkout ref "undefined"');

          case 20:
            _context3.next = 22;
            return resolveRef({ gitdir: gitdir, ref: ref });

          case 22:
            oid = _context3.sent;

          case 23:
            _context3.next = 25;
            return managers_js.GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 25:
            commit = _context3.sent;

            if (!(commit.type !== 'commit')) {
              _context3.next = 28;
              break;
            }

            throw new Error('Unexpected type: ' + commit.type);

          case 28:
            comm = models_js.GitCommit.from(commit.object.toString('utf8'));
            sha = comm.headers().tree;
            // Get top-level tree

            _context3.next = 32;
            return managers_js.GitObjectManager.read({ gitdir: gitdir, oid: sha });

          case 32:
            _ref6 = _context3.sent;
            type = _ref6.type;
            object = _ref6.object;

            if (!(type !== 'tree')) {
              _context3.next = 37;
              break;
            }

            throw new Error('Unexpected type: ' + type);

          case 37:
            tree = models_js.GitTree.from(object);
            // Acquire a lock on the index

            _context3.next = 40;
            return managers_js.GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref7 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(index) {
                var _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, entry;

                return _regeneratorRuntime.wrap(function _callee2$(_context2) {
                  while (1) {
                    switch (_context2.prev = _context2.next) {
                      case 0:
                        // TODO: Big optimization possible here.
                        // Instead of deleting and rewriting everything, only delete files
                        // that are not present in the new branch, and only write files that
                        // are not in the index or are in the index but have the wrong SHA.
                        _iteratorNormalCompletion2 = true;
                        _didIteratorError2 = false;
                        _iteratorError2 = undefined;
                        _context2.prev = 3;
                        _iterator2 = _getIterator(index);

                      case 5:
                        if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
                          _context2.next = 17;
                          break;
                        }

                        entry = _step2.value;
                        _context2.prev = 7;
                        _context2.next = 10;
                        return utils_js.rm(path.join(workdir, entry.path));

                      case 10:
                        _context2.next = 14;
                        break;

                      case 12:
                        _context2.prev = 12;
                        _context2.t0 = _context2['catch'](7);

                      case 14:
                        _iteratorNormalCompletion2 = true;
                        _context2.next = 5;
                        break;

                      case 17:
                        _context2.next = 23;
                        break;

                      case 19:
                        _context2.prev = 19;
                        _context2.t1 = _context2['catch'](3);
                        _didIteratorError2 = true;
                        _iteratorError2 = _context2.t1;

                      case 23:
                        _context2.prev = 23;
                        _context2.prev = 24;

                        if (!_iteratorNormalCompletion2 && _iterator2.return) {
                          _iterator2.return();
                        }

                      case 26:
                        _context2.prev = 26;

                        if (!_didIteratorError2) {
                          _context2.next = 29;
                          break;
                        }

                        throw _iteratorError2;

                      case 29:
                        return _context2.finish(26);

                      case 30:
                        return _context2.finish(23);

                      case 31:
                        index.clear();
                        // Write files. TODO: Write them atomically
                        _context2.next = 34;
                        return writeTreeToDisk({ gitdir: gitdir, workdir: workdir, index: index, prefix: '', tree: tree });

                      case 34:
                        // Update HEAD TODO: Handle non-branch cases
                        utils_js.write(gitdir + '/HEAD', 'ref: refs/heads/' + ref);

                      case 35:
                      case 'end':
                        return _context2.stop();
                    }
                  }
                }, _callee2, this, [[3, 19, 23, 31], [7, 12], [24,, 26, 30]]);
              }));

              return function (_x3) {
                return _ref7.apply(this, arguments);
              };
            }());

          case 40:
          case 'end':
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));

  return function checkout(_x2) {
    return _ref5.apply(this, arguments);
  };
}();

var clone = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var workdir = _ref.workdir,
        gitdir = _ref.gitdir,
        url = _ref.url,
        remote = _ref.remote,
        ref = _ref.ref,
        authUsername = _ref.authUsername,
        authPassword = _ref.authPassword,
        depth = _ref.depth,
        since = _ref.since,
        exclude = _ref.exclude,
        relative = _ref.relative,
        onprogress = _ref.onprogress;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            remote = remote || 'origin';
            _context.next = 3;
            return init({ gitdir: gitdir });

          case 3:
            _context.next = 5;
            return config({
              gitdir: gitdir,
              path: 'remote.' + remote + '.url',
              value: url
            });

          case 5:
            _context.next = 7;
            return fetch({
              gitdir: gitdir,
              ref: ref,
              remote: remote,
              authUsername: authUsername,
              authPassword: authPassword,
              depth: depth,
              since: since,
              exclude: exclude,
              relative: relative,
              onprogress: onprogress
            });

          case 7:
            _context.next = 9;
            return checkout({
              workdir: workdir,
              gitdir: gitdir,
              ref: ref,
              remote: remote
            });

          case 9:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function clone(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var constructTree = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        inode = _ref.inode;

    var children, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, _inode, entries, tree, oid;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            // use depth first traversal
            children = inode.children;
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context.prev = 4;
            _iterator = _getIterator(children);

          case 6:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context.next = 16;
              break;
            }

            _inode = _step.value;

            if (!(_inode.type === 'tree')) {
              _context.next = 13;
              break;
            }

            _inode.metadata.mode = '040000';
            _context.next = 12;
            return constructTree({ gitdir: gitdir, inode: _inode });

          case 12:
            _inode.metadata.oid = _context.sent;

          case 13:
            _iteratorNormalCompletion = true;
            _context.next = 6;
            break;

          case 16:
            _context.next = 22;
            break;

          case 18:
            _context.prev = 18;
            _context.t0 = _context['catch'](4);
            _didIteratorError = true;
            _iteratorError = _context.t0;

          case 22:
            _context.prev = 22;
            _context.prev = 23;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 25:
            _context.prev = 25;

            if (!_didIteratorError) {
              _context.next = 28;
              break;
            }

            throw _iteratorError;

          case 28:
            return _context.finish(25);

          case 29:
            return _context.finish(22);

          case 30:
            entries = children.map(function (inode) {
              return {
                mode: inode.metadata.mode,
                path: inode.basename,
                oid: inode.metadata.oid,
                type: inode.type
              };
            });
            tree = models_js.GitTree.from(entries);
            _context.next = 34;
            return managers_js.GitObjectManager.write({
              gitdir: gitdir,
              type: 'tree',
              object: tree.toObject()
            });

          case 34:
            oid = _context.sent;
            return _context.abrupt('return', oid);

          case 36:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[4, 18, 22, 30], [23,, 25, 29]]);
  }));

  return function constructTree(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var commit = function () {
  var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(_ref3) {
    var gitdir = _ref3.gitdir,
        author = _ref3.author,
        committer = _ref3.committer,
        message = _ref3.message,
        privateKeys = _ref3.privateKeys;
    var authorDateTime, committerDateTime, oid;
    return _regeneratorRuntime.wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            // Fill in missing arguments with default values
            if (author === undefined) author = {};

            if (!(author.name === undefined)) {
              _context3.next = 5;
              break;
            }

            _context3.next = 4;
            return config({ gitdir: gitdir, path: 'user.name' });

          case 4:
            author.name = _context3.sent;

          case 5:
            if (!(author.email === undefined)) {
              _context3.next = 9;
              break;
            }

            _context3.next = 8;
            return config({ gitdir: gitdir, path: 'user.email' });

          case 8:
            author.email = _context3.sent;

          case 9:
            committer = committer || author;
            authorDateTime = author.date || new Date();
            committerDateTime = committer.date || authorDateTime;
            oid = void 0;
            _context3.next = 15;
            return managers_js.GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref5 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(index) {
                var inode, treeRef, parents, parent, comm, branch;
                return _regeneratorRuntime.wrap(function _callee2$(_context2) {
                  while (1) {
                    switch (_context2.prev = _context2.next) {
                      case 0:
                        inode = utils_js.flatFileListToDirectoryStructure(index.entries);
                        _context2.next = 3;
                        return constructTree({ gitdir: gitdir, inode: inode });

                      case 3:
                        treeRef = _context2.sent;
                        parents = void 0;
                        _context2.prev = 5;
                        _context2.next = 8;
                        return resolveRef({ gitdir: gitdir, ref: 'HEAD' });

                      case 8:
                        parent = _context2.sent;

                        parents = [parent];
                        _context2.next = 15;
                        break;

                      case 12:
                        _context2.prev = 12;
                        _context2.t0 = _context2['catch'](5);

                        // Probably an initial commit
                        parents = [];

                      case 15:
                        comm = models_js.GitCommit.from({
                          tree: treeRef,
                          parent: parents,
                          author: {
                            name: author.name,
                            email: author.email,
                            timestamp: author.timestamp || Math.floor(authorDateTime.valueOf() / 1000),
                            timezoneOffset: author.timezoneOffset || 0
                          },
                          committer: {
                            name: committer.name,
                            email: committer.email,
                            timestamp: committer.timestamp || Math.floor(committerDateTime.valueOf() / 1000),
                            timezoneOffset: committer.timezoneOffset || 0
                          },
                          message: message
                        });

                        if (!privateKeys) {
                          _context2.next = 20;
                          break;
                        }

                        _context2.next = 19;
                        return comm.sign(privateKeys);

                      case 19:
                        comm = _context2.sent;

                      case 20:
                        _context2.next = 22;
                        return managers_js.GitObjectManager.write({
                          gitdir: gitdir,
                          type: 'commit',
                          object: comm.toObject()
                        });

                      case 22:
                        oid = _context2.sent;
                        _context2.next = 25;
                        return resolveRef({ gitdir: gitdir, ref: 'HEAD', depth: 2 });

                      case 25:
                        branch = _context2.sent;
                        _context2.next = 28;
                        return utils_js.write(path.join(gitdir, branch), oid + '\n');

                      case 28:
                      case 'end':
                        return _context2.stop();
                    }
                  }
                }, _callee2, this, [[5, 12]]);
              }));

              return function (_x3) {
                return _ref5.apply(this, arguments);
              };
            }());

          case 15:
            return _context3.abrupt('return', oid);

          case 16:
          case 'end':
            return _context3.stop();
        }
      }
    }, _callee3, this);
  }));

  return function commit(_x2) {
    return _ref4.apply(this, arguments);
  };
}();

var list = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref) {
    var gitdir = _ref.gitdir;
    var filenames;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            filenames = void 0;
            _context2.next = 3;
            return managers_js.GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(index) {
                return _regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        filenames = index.entries.map(function (x) {
                          return x.path;
                        });

                      case 1:
                      case 'end':
                        return _context.stop();
                    }
                  }
                }, _callee, this);
              }));

              return function (_x2) {
                return _ref3.apply(this, arguments);
              };
            }());

          case 3:
            return _context2.abrupt('return', filenames);

          case 4:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function list(_x) {
    return _ref2.apply(this, arguments);
  };
}();

// @flow
/*::
import type {Writable} from 'stream'
*/

var types$1 = {
  commit: 16,
  tree: 32,
  blob: 48,
  tag: 64
  // TODO: Move this to 'plumbing'
};var pack = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref /*: {oids: Array<string>, gitdir: string, outputStream: Writable} */
  ) {
    var oids = _ref.oids,
        gitdir = _ref.gitdir,
        outputStream = _ref.outputStream;

    var hash, stream$$1, write$$1, writeObject, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, oid, _ref4, type, object, digest;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            writeObject = function writeObject(_ref3) {
              var stype = _ref3.stype,
                  object = _ref3.object;

              var lastFour = void 0,
                  multibyte = void 0,
                  length = void 0;
              // Object type is encoded in bits 654
              var type = types$1[stype];
              if (type === undefined) throw new Error('Unrecognized type: ' + stype);
              // The length encoding get complicated.
              length = object.length;
              // Whether the next byte is part of the variable-length encoded number
              // is encoded in bit 7
              multibyte = length > 15 ? 128 : 0;
              // Last four bits of length is encoded in bits 3210
              lastFour = length & 15;
              // Discard those bits
              length = length >>> 4;
              // The first byte is then (1-bit multibyte?), (3-bit type), (4-bit least sig 4-bits of length)
              var byte = (multibyte | type | lastFour).toString(16);
              write$$1(byte, 'hex');
              // Now we keep chopping away at length 7-bits at a time until its zero,
              // writing out the bytes in what amounts to little-endian order.
              while (multibyte) {
                multibyte = length > 127 ? 128 : 0;
                byte = multibyte | length & 127;
                write$$1(pad(2, byte.toString(16), '0'), 'hex');
                length = length >>> 7;
              }
              // Lastly, we can compress and write the object.
              write$$1(buffer.Buffer.from(pako.deflate(object)));
            };

            write$$1 = function write$$1(chunk, enc) {
              stream$$1.write(chunk, enc);
              hash.update(chunk, enc);
            };

            hash = crypto.createHash('sha1');
            stream$$1 = outputStream;


            write$$1('PACK');
            write$$1('00000002', 'hex');
            // Write a 4 byte (32-bit) int
            write$$1(pad(8, oids.length.toString(16), '0'), 'hex');
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context.prev = 10;
            _iterator = _getIterator(oids);

          case 12:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context.next = 23;
              break;
            }

            oid = _step.value;
            _context.next = 16;
            return managers_js.GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 16:
            _ref4 = _context.sent;
            type = _ref4.type;
            object = _ref4.object;

            writeObject({ write: write$$1, object: object, stype: type });

          case 20:
            _iteratorNormalCompletion = true;
            _context.next = 12;
            break;

          case 23:
            _context.next = 29;
            break;

          case 25:
            _context.prev = 25;
            _context.t0 = _context['catch'](10);
            _didIteratorError = true;
            _iteratorError = _context.t0;

          case 29:
            _context.prev = 29;
            _context.prev = 30;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 32:
            _context.prev = 32;

            if (!_didIteratorError) {
              _context.next = 35;
              break;
            }

            throw _iteratorError;

          case 35:
            return _context.finish(32);

          case 36:
            return _context.finish(29);

          case 37:
            // Write SHA1 checksum
            digest = hash.digest();

            stream$$1.end(digest);
            return _context.abrupt('return', stream$$1);

          case 40:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[10, 25, 29, 37], [30,, 32, 36]]);
  }));

  return function pack(_x) {
    return _ref2.apply(this, arguments);
  };
}();

// @flow
var listCommits = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref /*: {
                                                                                gitdir: string,
                                                                                start: Array<string>,
                                                                                finish: Array<string>
                                                                                } */
  ) {
    /*: Set<string> */

    // Because git commits are named by their hash, there is no
    // way to construct a cycle. Therefore we won't worry about
    // setting a default recursion limit.
    var walk = function () {
      var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(oid) {
        var _ref5, type, object, commit, parents, _iteratorNormalCompletion3, _didIteratorError3, _iteratorError3, _iterator3, _step3;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                visited.add(oid);
                _context.next = 3;
                return managers_js.GitObjectManager.read({ gitdir: gitdir, oid: oid });

              case 3:
                _ref5 = _context.sent;
                type = _ref5.type;
                object = _ref5.object;

                if (!(type !== 'commit')) {
                  _context.next = 8;
                  break;
                }

                throw new Error('Expected type commit but type is ' + type);

              case 8:
                commit = models_js.GitCommit.from(object);
                parents = commit.headers().parent;
                _iteratorNormalCompletion3 = true;
                _didIteratorError3 = false;
                _iteratorError3 = undefined;
                _context.prev = 13;
                _iterator3 = _getIterator(parents);

              case 15:
                if (_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done) {
                  _context.next = 23;
                  break;
                }

                oid = _step3.value;

                if (!(!finishingSet.has(oid) && !visited.has(oid))) {
                  _context.next = 20;
                  break;
                }

                _context.next = 20;
                return walk(oid);

              case 20:
                _iteratorNormalCompletion3 = true;
                _context.next = 15;
                break;

              case 23:
                _context.next = 29;
                break;

              case 25:
                _context.prev = 25;
                _context.t0 = _context['catch'](13);
                _didIteratorError3 = true;
                _iteratorError3 = _context.t0;

              case 29:
                _context.prev = 29;
                _context.prev = 30;

                if (!_iteratorNormalCompletion3 && _iterator3.return) {
                  _iterator3.return();
                }

              case 32:
                _context.prev = 32;

                if (!_didIteratorError3) {
                  _context.next = 35;
                  break;
                }

                throw _iteratorError3;

              case 35:
                return _context.finish(32);

              case 36:
                return _context.finish(29);

              case 37:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this, [[13, 25, 29, 37], [30,, 32, 36]]);
      }));

      return function walk(_x2) {
        return _ref4.apply(this, arguments);
      };
    }();

    // Let's go walking!


    var gitdir = _ref.gitdir,
        start = _ref.start,
        finish = _ref.finish;

    var startingSet, finishingSet, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, ref, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, _ref3, _oid, visited, _iteratorNormalCompletion4, _didIteratorError4, _iteratorError4, _iterator4, _step4, oid;

    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            startingSet = new _Set();
            finishingSet = new _Set();
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context2.prev = 5;
            _iterator = _getIterator(start);

          case 7:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context2.next = 17;
              break;
            }

            ref = _step.value;
            _context2.t0 = startingSet;
            _context2.next = 12;
            return resolveRef({ gitdir: gitdir, ref: ref });

          case 12:
            _context2.t1 = _context2.sent;

            _context2.t0.add.call(_context2.t0, _context2.t1);

          case 14:
            _iteratorNormalCompletion = true;
            _context2.next = 7;
            break;

          case 17:
            _context2.next = 23;
            break;

          case 19:
            _context2.prev = 19;
            _context2.t2 = _context2['catch'](5);
            _didIteratorError = true;
            _iteratorError = _context2.t2;

          case 23:
            _context2.prev = 23;
            _context2.prev = 24;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 26:
            _context2.prev = 26;

            if (!_didIteratorError) {
              _context2.next = 29;
              break;
            }

            throw _iteratorError;

          case 29:
            return _context2.finish(26);

          case 30:
            return _context2.finish(23);

          case 31:
            _iteratorNormalCompletion2 = true;
            _didIteratorError2 = false;
            _iteratorError2 = undefined;
            _context2.prev = 34;
            _iterator2 = _getIterator(finish);

          case 36:
            if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
              _context2.next = 50;
              break;
            }

            _ref3 = _step2.value;
            _context2.prev = 38;
            _context2.next = 41;
            return resolveRef({ gitdir: gitdir, ref: _ref3 });

          case 41:
            _oid = _context2.sent;

            finishingSet.add(_oid);
            _context2.next = 47;
            break;

          case 45:
            _context2.prev = 45;
            _context2.t3 = _context2['catch'](38);

          case 47:
            _iteratorNormalCompletion2 = true;
            _context2.next = 36;
            break;

          case 50:
            _context2.next = 56;
            break;

          case 52:
            _context2.prev = 52;
            _context2.t4 = _context2['catch'](34);
            _didIteratorError2 = true;
            _iteratorError2 = _context2.t4;

          case 56:
            _context2.prev = 56;
            _context2.prev = 57;

            if (!_iteratorNormalCompletion2 && _iterator2.return) {
              _iterator2.return();
            }

          case 59:
            _context2.prev = 59;

            if (!_didIteratorError2) {
              _context2.next = 62;
              break;
            }

            throw _iteratorError2;

          case 62:
            return _context2.finish(59);

          case 63:
            return _context2.finish(56);

          case 64:
            visited = new _Set();
            _iteratorNormalCompletion4 = true;
            _didIteratorError4 = false;
            _iteratorError4 = undefined;
            _context2.prev = 68;
            _iterator4 = _getIterator(startingSet);

          case 70:
            if (_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done) {
              _context2.next = 77;
              break;
            }

            oid = _step4.value;
            _context2.next = 74;
            return walk(oid);

          case 74:
            _iteratorNormalCompletion4 = true;
            _context2.next = 70;
            break;

          case 77:
            _context2.next = 83;
            break;

          case 79:
            _context2.prev = 79;
            _context2.t5 = _context2['catch'](68);
            _didIteratorError4 = true;
            _iteratorError4 = _context2.t5;

          case 83:
            _context2.prev = 83;
            _context2.prev = 84;

            if (!_iteratorNormalCompletion4 && _iterator4.return) {
              _iterator4.return();
            }

          case 86:
            _context2.prev = 86;

            if (!_didIteratorError4) {
              _context2.next = 89;
              break;
            }

            throw _iteratorError4;

          case 89:
            return _context2.finish(86);

          case 90:
            return _context2.finish(83);

          case 91:
            return _context2.abrupt('return', visited);

          case 92:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this, [[5, 19, 23, 31], [24,, 26, 30], [34, 52, 56, 64], [38, 45], [57,, 59, 63], [68, 79, 83, 91], [84,, 86, 90]]);
  }));

  return function listCommits(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var listObjects = function () {
  var _ref7 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee4(_ref6 /*: {
                                                                                 gitdir: string,
                                                                                 oids: Array<string>
                                                                                 } */
  ) {

    // We don't do the purest simplest recursion, because we can
    // avoid reading Blob objects entirely since the Tree objects
    // tell us which oids are Blobs and which are Trees.
    var walk = function () {
      var _ref8 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(oid) {
        var _ref9, type, object, commit, tree, _tree, _iteratorNormalCompletion5, _didIteratorError5, _iteratorError5, _iterator5, _step5, entry;

        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                visited.add(oid);
                _context3.next = 3;
                return managers_js.GitObjectManager.read({ gitdir: gitdir, oid: oid });

              case 3:
                _ref9 = _context3.sent;
                type = _ref9.type;
                object = _ref9.object;

                if (!(type === 'commit')) {
                  _context3.next = 13;
                  break;
                }

                commit = models_js.GitCommit.from(object);
                tree = commit.headers().tree;
                _context3.next = 11;
                return walk(tree);

              case 11:
                _context3.next = 43;
                break;

              case 13:
                if (!(type === 'tree')) {
                  _context3.next = 43;
                  break;
                }

                _tree = models_js.GitTree.from(object);
                _iteratorNormalCompletion5 = true;
                _didIteratorError5 = false;
                _iteratorError5 = undefined;
                _context3.prev = 18;
                _iterator5 = _getIterator( /*: TreeEntry */_tree);

              case 20:
                if (_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done) {
                  _context3.next = 29;
                  break;
                }

                entry = _step5.value;

                visited.add(entry.oid);
                // only recurse for trees

                if (!(entry.type === 'tree')) {
                  _context3.next = 26;
                  break;
                }

                _context3.next = 26;
                return walk(entry.oid);

              case 26:
                _iteratorNormalCompletion5 = true;
                _context3.next = 20;
                break;

              case 29:
                _context3.next = 35;
                break;

              case 31:
                _context3.prev = 31;
                _context3.t0 = _context3['catch'](18);
                _didIteratorError5 = true;
                _iteratorError5 = _context3.t0;

              case 35:
                _context3.prev = 35;
                _context3.prev = 36;

                if (!_iteratorNormalCompletion5 && _iterator5.return) {
                  _iterator5.return();
                }

              case 38:
                _context3.prev = 38;

                if (!_didIteratorError5) {
                  _context3.next = 41;
                  break;
                }

                throw _iteratorError5;

              case 41:
                return _context3.finish(38);

              case 42:
                return _context3.finish(35);

              case 43:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, this, [[18, 31, 35, 43], [36,, 38, 42]]);
      }));

      return function walk(_x4) {
        return _ref8.apply(this, arguments);
      };
    }();

    // Let's go walking!


    var gitdir = _ref6.gitdir,
        oids = _ref6.oids;

    var visited, _iteratorNormalCompletion6, _didIteratorError6, _iteratorError6, _iterator6, _step6, oid;

    return _regeneratorRuntime.wrap(function _callee4$(_context4) {
      while (1) {
        switch (_context4.prev = _context4.next) {
          case 0:
            visited /*: Set<string> */ = new _Set();
            _iteratorNormalCompletion6 = true;
            _didIteratorError6 = false;
            _iteratorError6 = undefined;
            _context4.prev = 4;
            _iterator6 = _getIterator(oids);

          case 6:
            if (_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done) {
              _context4.next = 13;
              break;
            }

            oid = _step6.value;
            _context4.next = 10;
            return walk(oid);

          case 10:
            _iteratorNormalCompletion6 = true;
            _context4.next = 6;
            break;

          case 13:
            _context4.next = 19;
            break;

          case 15:
            _context4.prev = 15;
            _context4.t0 = _context4['catch'](4);
            _didIteratorError6 = true;
            _iteratorError6 = _context4.t0;

          case 19:
            _context4.prev = 19;
            _context4.prev = 20;

            if (!_iteratorNormalCompletion6 && _iterator6.return) {
              _iterator6.return();
            }

          case 22:
            _context4.prev = 22;

            if (!_didIteratorError6) {
              _context4.next = 25;
              break;
            }

            throw _iteratorError6;

          case 25:
            return _context4.finish(22);

          case 26:
            return _context4.finish(19);

          case 27:
            return _context4.abrupt('return', visited);

          case 28:
          case 'end':
            return _context4.stop();
        }
      }
    }, _callee4, this, [[4, 15, 19, 27], [20,, 22, 26]]);
  }));

  return function listObjects(_x3) {
    return _ref7.apply(this, arguments);
  };
}();

var push = function () {
  var _ref11 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee5(_ref10) {
    var gitdir = _ref10.gitdir,
        ref = _ref10.ref,
        remote = _ref10.remote,
        url = _ref10.url,
        authUsername = _ref10.authUsername,
        authPassword = _ref10.authPassword;
    var fullRef, oid, httpRemote, commits, objects, packstream, oldoid, response;
    return _regeneratorRuntime.wrap(function _callee5$(_context5) {
      while (1) {
        switch (_context5.prev = _context5.next) {
          case 0:
            // TODO: Figure out how pushing tags works. (This only works for branches.)
            remote = remote || 'origin';

            if (!(url === undefined)) {
              _context5.next = 5;
              break;
            }

            _context5.next = 4;
            return config({
              gitdir: gitdir,
              path: 'remote.' + remote + '.url'
            });

          case 4:
            url = _context5.sent;

          case 5:
            fullRef = ref.startsWith('refs/') ? ref : 'refs/heads/' + ref;
            _context5.next = 8;
            return resolveRef({ gitdir: gitdir, ref: ref });

          case 8:
            oid = _context5.sent;
            httpRemote = new managers_js.GitRemoteHTTP(url);

            if (authUsername !== undefined && authPassword !== undefined) {
              httpRemote.auth = {
                username: authUsername,
                password: authPassword
              };
            }
            _context5.next = 13;
            return httpRemote.preparePush();

          case 13:
            _context5.next = 15;
            return listCommits({
              gitdir: gitdir,
              start: [oid],
              finish: httpRemote.refs.values()
            });

          case 15:
            commits = _context5.sent;
            _context5.next = 18;
            return listObjects({ gitdir: gitdir, oids: commits });

          case 18:
            objects = _context5.sent;
            packstream = new stream.PassThrough();
            oldoid = httpRemote.refs.get(fullRef) || '0000000000000000000000000000000000000000';

            packstream.write(models_js.GitPktLine.encode(oldoid + ' ' + oid + ' ' + fullRef + '\0 report-status\n'));
            packstream.write(models_js.GitPktLine.flush());
            pack({
              gitdir: gitdir,
              oids: [].concat(_toConsumableArray(objects)),
              outputStream: packstream
            });
            _context5.next = 26;
            return httpRemote.push(packstream);

          case 26:
            response = _context5.sent;
            return _context5.abrupt('return', response);

          case 28:
          case 'end':
            return _context5.stop();
        }
      }
    }, _callee5, this);
  }));

  return function push(_x5) {
    return _ref11.apply(this, arguments);
  };
}();

// @flow
var log = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        _ref$ref = _ref.ref,
        ref = _ref$ref === undefined ? 'HEAD' : _ref$ref,
        depth = _ref.depth,
        since = _ref.since;

    var sinceTimestamp, commits, start, _ref3, type, object, currentCommit, oid, _ref4, _type, _object;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            sinceTimestamp = since === undefined ? undefined : Math.floor(since.valueOf() / 1000);
            // TODO: In the future, we may want to have an API where we return a
            // async iterator that emits commits.

            commits = [];
            _context.next = 4;
            return resolveRef({ gitdir: gitdir, ref: ref });

          case 4:
            start = _context.sent;
            _context.next = 7;
            return managers_js.GitObjectManager.read({ gitdir: gitdir, oid: start });

          case 7:
            _ref3 = _context.sent;
            type = _ref3.type;
            object = _ref3.object;

            if (!(type !== 'commit')) {
              _context.next = 12;
              break;
            }

            throw new Error('The given ref ' + ref + ' did not resolve to a commit but to a ' + type);

          case 12:
            currentCommit = _extends({ oid: start }, models_js.GitCommit.from(object).parse());

            commits.push(currentCommit);

          case 14:
            

            if (!(depth !== undefined && commits.length === depth)) {
              _context.next = 17;
              break;
            }

            return _context.abrupt('break', 33);

          case 17:
            if (!(currentCommit.parent.length === 0)) {
              _context.next = 19;
              break;
            }

            return _context.abrupt('break', 33);

          case 19:
            oid = currentCommit.parent[0];
            _context.next = 22;
            return managers_js.GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 22:
            _ref4 = _context.sent;
            _type = _ref4.type;
            _object = _ref4.object;

            if (!(_type !== 'commit')) {
              _context.next = 27;
              break;
            }

            throw new Error('Invalid commit parent ' + currentCommit.parent[0] + ' is of type ' + _type);

          case 27:
            currentCommit = _extends({ oid: oid }, models_js.GitCommit.from(_object).parse());

            if (!(sinceTimestamp !== undefined && currentCommit.author.timestamp <= sinceTimestamp)) {
              _context.next = 30;
              break;
            }

            return _context.abrupt('break', 33);

          case 30:
            commits.push(currentCommit);
            _context.next = 14;
            break;

          case 33:
            return _context.abrupt('return', commits);

          case 34:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function log(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var remove = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref) {
    var gitdir = _ref.gitdir,
        filepath = _ref.filepath;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            _context2.next = 2;
            return managers_js.GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(index) {
                return _regeneratorRuntime.wrap(function _callee$(_context) {
                  while (1) {
                    switch (_context.prev = _context.next) {
                      case 0:
                        index.delete({ filepath: filepath });

                      case 1:
                      case 'end':
                        return _context.stop();
                    }
                  }
                }, _callee, this);
              }));

              return function (_x2) {
                return _ref3.apply(this, arguments);
              };
            }());

          case 2:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function remove(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var HttpKeyServer = new openpgp_min_js.HKP();

var verify = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir,
        ref = _ref.ref,
        publicKeys = _ref.publicKeys;

    var oid, _ref3, type, object, commit, author, keys, keyArray, validity;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.next = 2;
            return resolveRef({ gitdir: gitdir, ref: ref });

          case 2:
            oid = _context.sent;
            _context.next = 5;
            return managers_js.GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 5:
            _ref3 = _context.sent;
            type = _ref3.type;
            object = _ref3.object;

            if (!(type !== 'commit')) {
              _context.next = 10;
              break;
            }

            throw new Error('git.verify() was expecting a ref type \'commit\' but got type \'' + type + '\'');

          case 10:
            commit = models_js.GitCommit.from(object);
            author = commit.headers().author;
            _context.next = 14;
            return commit.listSigningKeys();

          case 14:
            keys = _context.sent;

            if (publicKeys) {
              _context.next = 20;
              break;
            }

            _context.next = 18;
            return _Promise.all(keys.map(function (id) {
              return HttpKeyServer.lookup({ keyId: id });
            }));

          case 18:
            keyArray = _context.sent;

            publicKeys = keyArray.join('\n');

          case 20:
            _context.next = 22;
            return commit.verify(publicKeys);

          case 22:
            validity = _context.sent;

            if (validity) {
              _context.next = 25;
              break;
            }

            return _context.abrupt('return', false);

          case 25:
            return _context.abrupt('return', { author: author, keys: keys });

          case 26:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function verify(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var getOidAtPath = function () {
  var _ref3 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref2 /*: {
                                                                                gitdir: string,
                                                                                tree: GitTree,
                                                                                path: string|Array<string>
                                                                                } */
  ) {
    var gitdir = _ref2.gitdir,
        tree = _ref2.tree,
        path$$1 = _ref2.path;

    var dirname, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, entry, _ref4, type, object, _tree;

    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            if (typeof path$$1 === 'string') path$$1 = path$$1.split('/');
            dirname = path$$1.shift();
            _iteratorNormalCompletion = true;
            _didIteratorError = false;
            _iteratorError = undefined;
            _context.prev = 5;
            _iterator = _getIterator(tree);

          case 7:
            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
              _context.next = 25;
              break;
            }

            entry = _step.value;

            if (!(entry.path === dirname)) {
              _context.next = 22;
              break;
            }

            if (!(path$$1.length === 0)) {
              _context.next = 12;
              break;
            }

            return _context.abrupt('return', entry.oid);

          case 12:
            _context.next = 14;
            return managers_js.GitObjectManager.read({
              gitdir: gitdir,
              oid: entry.oid
            });

          case 14:
            _ref4 = _context.sent;
            type = _ref4.type;
            object = _ref4.object;

            if (!(type === 'tree')) {
              _context.next = 20;
              break;
            }

            _tree = models_js.GitTree.from(object);
            return _context.abrupt('return', getOidAtPath({ gitdir: gitdir, tree: _tree, path: path$$1 }));

          case 20:
            if (!(type === 'blob')) {
              _context.next = 22;
              break;
            }

            throw new Error('Blob found where tree expected.');

          case 22:
            _iteratorNormalCompletion = true;
            _context.next = 7;
            break;

          case 25:
            _context.next = 31;
            break;

          case 27:
            _context.prev = 27;
            _context.t0 = _context['catch'](5);
            _didIteratorError = true;
            _iteratorError = _context.t0;

          case 31:
            _context.prev = 31;
            _context.prev = 32;

            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }

          case 34:
            _context.prev = 34;

            if (!_didIteratorError) {
              _context.next = 37;
              break;
            }

            throw _iteratorError;

          case 37:
            return _context.finish(34);

          case 38:
            return _context.finish(31);

          case 39:
            return _context.abrupt('return', null);

          case 40:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[5, 27, 31, 39], [32,, 34, 38]]);
  }));

  return function getOidAtPath(_x) {
    return _ref3.apply(this, arguments);
  };
}();

var getHeadTree = function () {
  var _ref6 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(_ref5) {
    var gitdir = _ref5.gitdir;

    var oid, _ref7, cobject, commit, _ref8, tobject, tree;

    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            _context2.next = 2;
            return resolveRef({ gitdir: gitdir, ref: 'HEAD' });

          case 2:
            oid = _context2.sent;
            _context2.next = 5;
            return managers_js.GitObjectManager.read({ gitdir: gitdir, oid: oid });

          case 5:
            _ref7 = _context2.sent;
            cobject = _ref7.object;
            commit = models_js.GitCommit.from(cobject);
            _context2.next = 10;
            return managers_js.GitObjectManager.read({
              gitdir: gitdir,
              oid: commit.parseHeaders().tree
            });

          case 10:
            _ref8 = _context2.sent;
            tobject = _ref8.object;
            tree = models_js.GitTree.from(tobject).entries();
            return _context2.abrupt('return', tree);

          case 14:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function getHeadTree(_x2) {
    return _ref6.apply(this, arguments);
  };
}();

// For now we're just diffing individual files


// @flow
var status = function () {
  var _ref10 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee4(_ref9 /*: {
                                                                                  workdir: string,
                                                                                  gitdir: string,
                                                                                  pathname: string
                                                                                  } */
  ) {
    var workdir = _ref9.workdir,
        gitdir = _ref9.gitdir,
        pathname = _ref9.pathname;

    var ignored, headTree, treeOid, indexEntry, stats, object, workdirOid, _object, _workdirOid;

    return _regeneratorRuntime.wrap(function _callee4$(_context4) {
      while (1) {
        switch (_context4.prev = _context4.next) {
          case 0:
            _context4.next = 2;
            return managers_js.GitIgnoreManager.isIgnored({ gitdir: gitdir, workdir: workdir, pathname: pathname });

          case 2:
            ignored = _context4.sent;

            if (!ignored) {
              _context4.next = 5;
              break;
            }

            return _context4.abrupt('return', 'ignored');

          case 5:
            _context4.next = 7;
            return getHeadTree({ gitdir: gitdir });

          case 7:
            headTree = _context4.sent;
            _context4.next = 10;
            return getOidAtPath({ gitdir: gitdir, tree: headTree, path: pathname });

          case 10:
            treeOid = _context4.sent;
            indexEntry = null;
            // Acquire a lock on the index

            _context4.next = 14;
            return managers_js.GitIndexManager.acquire(gitdir + '/index', function () {
              var _ref11 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee3(index) {
                var _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, entry;

                return _regeneratorRuntime.wrap(function _callee3$(_context3) {
                  while (1) {
                    switch (_context3.prev = _context3.next) {
                      case 0:
                        _iteratorNormalCompletion2 = true;
                        _didIteratorError2 = false;
                        _iteratorError2 = undefined;
                        _context3.prev = 3;
                        _iterator2 = _getIterator(index);

                      case 5:
                        if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
                          _context3.next = 13;
                          break;
                        }

                        entry = _step2.value;

                        if (!(entry.path === pathname)) {
                          _context3.next = 10;
                          break;
                        }

                        indexEntry = entry;
                        return _context3.abrupt('break', 13);

                      case 10:
                        _iteratorNormalCompletion2 = true;
                        _context3.next = 5;
                        break;

                      case 13:
                        _context3.next = 19;
                        break;

                      case 15:
                        _context3.prev = 15;
                        _context3.t0 = _context3['catch'](3);
                        _didIteratorError2 = true;
                        _iteratorError2 = _context3.t0;

                      case 19:
                        _context3.prev = 19;
                        _context3.prev = 20;

                        if (!_iteratorNormalCompletion2 && _iterator2.return) {
                          _iterator2.return();
                        }

                      case 22:
                        _context3.prev = 22;

                        if (!_didIteratorError2) {
                          _context3.next = 25;
                          break;
                        }

                        throw _iteratorError2;

                      case 25:
                        return _context3.finish(22);

                      case 26:
                        return _context3.finish(19);

                      case 27:
                      case 'end':
                        return _context3.stop();
                    }
                  }
                }, _callee3, this, [[3, 15, 19, 27], [20,, 22, 26]]);
              }));

              return function (_x4) {
                return _ref11.apply(this, arguments);
              };
            }());

          case 14:
            stats = null;
            _context4.prev = 15;
            _context4.next = 18;
            return pify(utils_js.fs().lstat)(path.join(workdir, pathname));

          case 18:
            stats = _context4.sent;
            _context4.next = 25;
            break;

          case 21:
            _context4.prev = 21;
            _context4.t0 = _context4['catch'](15);

            if (!(_context4.t0.code !== 'ENOENT')) {
              _context4.next = 25;
              break;
            }

            throw _context4.t0;

          case 25:
            if (!(treeOid === null && stats === null)) {
              _context4.next = 33;
              break;
            }

            if (!(indexEntry === null)) {
              _context4.next = 30;
              break;
            }

            return _context4.abrupt('return', 'absent');

          case 30:
            return _context4.abrupt('return', '*absent');

          case 31:
            _context4.next = 68;
            break;

          case 33:
            if (!(treeOid !== null && stats === null)) {
              _context4.next = 41;
              break;
            }

            if (!(indexEntry === null)) {
              _context4.next = 38;
              break;
            }

            return _context4.abrupt('return', 'deleted');

          case 38:
            return _context4.abrupt('return', '*deleted');

          case 39:
            _context4.next = 68;
            break;

          case 41:
            if (!(treeOid === null && stats !== null)) {
              _context4.next = 55;
              break;
            }

            if (!(indexEntry === null)) {
              _context4.next = 46;
              break;
            }

            return _context4.abrupt('return', '*added');

          case 46:
            _context4.next = 48;
            return utils_js.read(path.join(workdir, pathname));

          case 48:
            object = _context4.sent;
            _context4.next = 51;
            return managers_js.GitObjectManager.hash({
              gitdir: gitdir,
              type: 'blob',
              object: object
            });

          case 51:
            workdirOid = _context4.sent;
            return _context4.abrupt('return', workdirOid === indexEntry.oid ? 'added' : '*added');

          case 53:
            _context4.next = 68;
            break;

          case 55:
            _context4.next = 57;
            return utils_js.read(path.join(workdir, pathname));

          case 57:
            _object = _context4.sent;
            _context4.next = 60;
            return managers_js.GitObjectManager.hash({
              gitdir: gitdir,
              type: 'blob',
              object: _object
            });

          case 60:
            _workdirOid = _context4.sent;

            if (!(_workdirOid === treeOid)) {
              _context4.next = 65;
              break;
            }

            if (!(indexEntry === null)) {
              _context4.next = 64;
              break;
            }

            return _context4.abrupt('return', '*unmodified');

          case 64:
            return _context4.abrupt('return', _workdirOid === indexEntry.oid ? 'unmodified' : '*unmodified');

          case 65:
            if (!(indexEntry === null)) {
              _context4.next = 67;
              break;
            }

            return _context4.abrupt('return', '*modified');

          case 67:
            return _context4.abrupt('return', _workdirOid === indexEntry.oid ? 'modified' : '*modified');

          case 68:
          case 'end':
            return _context4.stop();
        }
      }
    }, _callee4, this, [[15, 21]]);
  }));

  return function status(_x3) {
    return _ref10.apply(this, arguments);
  };
}();

var test = function () {
  var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(filepath) {
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.prev = 0;
            _context.next = 3;
            return pify(utils_js.fs().lstat)(path.join(filepath, '.git'));

          case 3:
            return _context.abrupt('return', true);

          case 6:
            _context.prev = 6;
            _context.t0 = _context['catch'](0);
            return _context.abrupt('return', false);

          case 9:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this, [[0, 6]]);
  }));

  return function test(_x) {
    return _ref.apply(this, arguments);
  };
}();
// TODO: Detect base repositories?


var findRoot = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(filepath) {
    var parent;
    return _regeneratorRuntime.wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            _context2.next = 2;
            return test(filepath);

          case 2:
            if (!_context2.sent) {
              _context2.next = 4;
              break;
            }

            return _context2.abrupt('return', filepath);

          case 4:
            parent = path.dirname(filepath);

            if (!(parent === filepath)) {
              _context2.next = 7;
              break;
            }

            throw new Error('Unable to find git root');

          case 7:
            return _context2.abrupt('return', findRoot(parent));

          case 8:
          case 'end':
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function findRoot(_x2) {
    return _ref2.apply(this, arguments);
  };
}();

// Note: this function will get more interesting once we add support for
// packed-refs.
var listBranches = function () {
  var _ref2 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(_ref) {
    var gitdir = _ref.gitdir;
    var files;
    return _regeneratorRuntime.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.next = 2;
            return pify(utils_js.fs().readdir)(gitdir + '/refs/heads');

          case 2:
            files = _context.sent;
            return _context.abrupt('return', files);

          case 4:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  }));

  return function listBranches(_x) {
    return _ref2.apply(this, arguments);
  };
}();

var version$1 = "0.0.20";

function version() {
  return version$1;
}

exports.add = add;
exports.clone = clone;
exports.checkout = checkout;
exports.commit = commit;
exports.fetch = fetch;
exports.init = init;
exports.list = list;
exports.listCommits = listCommits;
exports.listObjects = listObjects;
exports.log = log;
exports.pack = pack;
exports.push = push;
exports.remove = remove;
exports.resolveRef = resolveRef;
exports.config = config;
exports.unpack = unpack;
exports.verify = verify;
exports.status = status;
exports.findRoot = findRoot;
exports.listBranches = listBranches;
exports.version = version;
//# sourceMappingURL=commands.js.map
